"""A client for interacting with the rippled WebSocket API."""
from __future__ import annotations

import json
from asyncio import Future, Queue, Task, create_task, get_running_loop
from random import randrange
from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from typing_extensions import Final
from websockets.legacy.client import WebSocketClientProtocol, connect

from xrpl.asyncio.clients.client import Client
from xrpl.asyncio.clients.exceptions import XRPLWebsocketException
from xrpl.asyncio.clients.utils import request_to_websocket, websocket_to_response
from xrpl.models.requests.request import Request
from xrpl.models.response import Response

_REQ_ID_MAX: Final[int] = 1_000_000
# the types from asyncio are not implemented as generics in python 3.8 and
# lower, so we need to only subscript them when running typechecking.
if TYPE_CHECKING:
    _REQUESTS_TYPE = Dict[str, Future[Dict[str, Any]]]
    _MESSAGES_TYPE = Queue[Dict[str, Any]]
    _HANDLER_TYPE = Task[None]
else:
    _REQUESTS_TYPE = Dict[str, Future]
    _MESSAGES_TYPE = Queue
    _HANDLER_TYPE = Task


def _inject_request_id(request: Request) -> Request:
    """
    Given a Request with an ID, return the same Request.

    Given a Request without an ID, make a copy with a randomly generated ID.
    """
    if request.id is not None:
        return request
    request_dict = request.to_dict()
    request_dict["id"] = f"{request.method}_{randrange(_REQ_ID_MAX)}"
    return Request.from_dict(request_dict)


class WebsocketBase(Client):
    """
    A client for interacting with the rippled WebSocket API.

    :meta private:
    """

    def __init__(self: WebsocketBase, url: str) -> None:
        """
        Initializes a websocket client.

        Arguments:
            url: The URL of the rippled node to submit requests to.
        """
        self._open_requests: _REQUESTS_TYPE = {}
        self._websocket: Optional[WebSocketClientProtocol] = None
        self._handler_task: Optional[_HANDLER_TYPE] = None
        # unfortunately, we cannot create the Queue here because it needs to be
        # tied to a currently-running event loop. the sync websocket client
        # will initialize a new event loop when it opens the connection, so for
        # that client the initializer cannot create the queue
        self._messages: Optional[_MESSAGES_TYPE] = None
        super().__init__(url)

    def is_open(self: WebsocketBase) -> bool:
        """
        Returns whether the client is currently open.

        Returns:
            True if the client is currently open, False otherwise.
        """
        return (
            self._handler_task is not None
            and self._messages is not None
            and self._websocket is not None
            and self._websocket.open
        )

    async def _do_open(self: WebsocketBase) -> None:
        """Connects the client to the Web Socket API at its URL."""
        # open the connection
        self._websocket = await connect(self.url)

        # make a message queue
        self._messages = Queue()

        # start the handler
        self._handler_task = create_task(self._handler())

    async def _do_close(self: WebsocketBase) -> None:
        """Closes the connection."""
        # cancel the handler
        cast(_HANDLER_TYPE, self._handler_task).cancel()
        self._handler_task = None

        # cancel any pending request Futures
        for future in self._open_requests.values():
            future.cancel()
        self._open_requests = {}

        # clear the message queue
        for _ in range(cast(_MESSAGES_TYPE, self._messages).qsize()):
            cast(_MESSAGES_TYPE, self._messages).get_nowait()
            cast(_MESSAGES_TYPE, self._messages).task_done()
        self._messages = None

        # close the connection
        await cast(WebSocketClientProtocol, self._websocket).close()

    async def _handler(self: WebsocketBase) -> None:
        """
        This is basically a middleware for the websocket library. For all received
        messages we check whether there is an outstanding future we need to resolve,
        and if so do so.

        Then we store the already-parsed JSON in our own queue for generic iteration.

        As long as a given client remains open, this handler will be running as a Task.
        """
        async for response in cast(WebSocketClientProtocol, self._websocket):
            response_dict = json.loads(response)

            # if this response corresponds to request, fulfill the Future
            if "id" in response_dict and response_dict["id"] in self._open_requests:
                self._open_requests[response_dict["id"]].set_result(response_dict)

            # enqueue the response for the message queue
            cast(_MESSAGES_TYPE, self._messages).put_nowait(response_dict)

    def _set_up_future(self: WebsocketBase, request: Request) -> None:
        """
        Only to be called from the public send and _request_impl functions.
        Given a request with an ID, ensure that that ID is backed by an open
        Future in self._open_requests.
        """
        if request.id is None:
            return
        request_str = str(request.id)
        if (
            request_str in self._open_requests
            and not self._open_requests[request_str].done()
        ):
            raise XRPLWebsocketException(
                f"Request {request_str} is already in progress."
            )
        self._open_requests[request_str] = get_running_loop().create_future()

    async def _do_send_no_future(self: WebsocketBase, request: Request) -> None:
        await cast(WebSocketClientProtocol, self._websocket).send(
            json.dumps(
                request_to_websocket(request),
            ),
        )

    async def _do_send(self: WebsocketBase, request: Request) -> None:
        # we need to set up a future here, even if no one cares about it, so
        # that if a user submits a few requests with the same ID they fail.
        self._set_up_future(request)
        await self._do_send_no_future(request)

    async def _do_pop_message(self: WebsocketBase) -> Dict[str, Any]:
        msg = await cast(_MESSAGES_TYPE, self._messages).get()
        cast(_MESSAGES_TYPE, self._messages).task_done()
        return msg

    async def _do_request_impl(self: WebsocketBase, request: Request) -> Response:
        """
        Base ``_request_impl`` implementation for websockets.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketBase is not open.
        """
        # if no ID on this request, generate and inject one, and ensure it
        # is backed by a future
        request_with_id = _inject_request_id(request)
        request_str = str(request_with_id.id)
        self._set_up_future(request_with_id)

        # fire-and-forget the send, and await the Future
        create_task(self._do_send_no_future(request_with_id))
        raw_response = await self._open_requests[request_str]

        # remove the resolved Future, hopefully getting it garbage colleted
        del self._open_requests[request_str]
        return websocket_to_response(raw_response)



"""Helper functions for balance parser."""

from decimal import Decimal
from typing import Callable, List, Optional

from xrpl.models.transactions.metadata import TransactionMetadata
from xrpl.utils.txn_parser.utils.nodes import NormalizedNode, normalize_nodes
from xrpl.utils.txn_parser.utils.parser import group_by_account
from xrpl.utils.txn_parser.utils.types import AccountBalance, AccountBalances, Balance
from xrpl.utils.xrp_conversions import drops_to_xrp


def _get_xrp_quantity(
    node: NormalizedNode,
    value: Optional[Decimal],
) -> Optional[AccountBalance]:
    if value is None:
        return None
    absolute_value = value.copy_abs()
    xrp_value = (
        drops_to_xrp(str(absolute_value)).copy_negate()
        if value.is_signed()
        else drops_to_xrp(str(absolute_value))
    )
    final_fields = node.get("FinalFields")
    if final_fields is not None:
        account = final_fields.get("Account")
        if account is not None:
            return AccountBalance(
                account=account,
                balance=Balance(
                    currency="XRP",
                    value=f"{xrp_value.normalize():f}",
                ),
            )
    new_fields = node.get("NewFields")
    if new_fields is not None:
        account = new_fields.get("Account")
        if account is not None:
            return AccountBalance(
                account=account,
                balance=Balance(
                    currency="XRP",
                    value=f"{xrp_value.normalize():f}",
                ),
            )
    return None


def _flip_trustline_perspective(account_balance: AccountBalance) -> AccountBalance:
    balance = account_balance["balance"]
    negated_value = Decimal(balance["value"]).copy_negate()
    issuer = balance["issuer"]
    return AccountBalance(
        account=issuer,
        balance=Balance(
            currency=balance["currency"],
            issuer=account_balance["account"],
            value=f"{negated_value.normalize():f}",
        ),
    )


def _get_trustline_quantity(
    node: NormalizedNode,
    value: Optional[Decimal],
) -> List[AccountBalance]:
    """
    Computes the complete list of every balance affected by the transaction.

    Args:
        node: The affected node.
        value: The currency amount value.

    Returns:
        A list of computed balances.
    """
    if value is None:
        return []
    fields = (
        node["NewFields"] if node.get("NewFields") is not None else node["FinalFields"]
    )
    assert fields is not None
    low_limit = fields.get("LowLimit")
    balance = fields.get("Balance")
    high_limit = fields.get("HighLimit")
    if low_limit is not None and balance is not None and high_limit is not None:
        low_limit_issuer = low_limit.get("issuer")
        assert isinstance(balance, dict)
        balance_currency = balance.get("currency")
        high_limit_issuer = high_limit.get("issuer")
        if (
            low_limit_issuer is not None
            and balance_currency is not None
            and high_limit_issuer is not None
        ):
            result = AccountBalance(
                account=low_limit_issuer,
                balance=Balance(
                    currency=balance_currency,
                    issuer=high_limit_issuer,
                    value=f"{value.normalize():f}",
                ),
            )
            return [result, _flip_trustline_perspective(result)]
    return []


def _get_node_balances(
    node: NormalizedNode,
    value: Optional[Decimal],
) -> List[AccountBalance]:
    """
    Retrieve the balance from a node.

    Args:
        node: The affected node.
        value: The currency amount's value

    Returns:
        A list of balances.
    """
    if node["LedgerEntryType"] == "AccountRoot":
        xrp_quantity = _get_xrp_quantity(node, value)
        if xrp_quantity is None:
            return []
        return [xrp_quantity]
    if node["LedgerEntryType"] == "RippleState":
        trustline_quantity = _get_trustline_quantity(node, value)
        return trustline_quantity
    return []


def _group_balances_by_account(
    account_balances: List[AccountBalance],
) -> List[AccountBalances]:
    grouped_balances = group_by_account(account_balances)
    result = []
    for account, account_obj in grouped_balances.items():
        balances: List[Balance] = [balance["balance"] for balance in account_obj]
        result.append(
            AccountBalances(
                account=account,
                balances=balances,
            )
        )
    return result


def derive_account_balances(
    metadata: TransactionMetadata,
    parser: Callable[[NormalizedNode], Optional[Decimal]],
) -> List[AccountBalances]:
    """
    Derives the account balances from a node.

    Args:
        metadata: Transactions metadata.
        parser: The balance parser.

    Returns:
        All balances affected by the transaction.
        The balances are grouped by their accounts.
    """
    quantities = [
        quantity
        for node in normalize_nodes(metadata)
        for quantity in _get_node_balances(node, parser(node))
    ]
    return _group_balances_by_account(quantities)



"""Helper functions for the clients module."""

from typing import Any, Dict

from xrpl.models.requests.request import Request
from xrpl.models.response import Response, ResponseStatus, ResponseType


def request_to_json_rpc(request_object: Request) -> Dict[str, Any]:
    """Converts a request model object to the appropriate JSON format for
    interacting with the rippled API.

    Args:
        request_object: A Request object representing the parameters of a
                        request to the rippled JSON RPC.

    Returns:
        A dictionary containing the attributes of this Request object formatted
        for submission to the rippled JSON RPC.
    """
    params = request_object.to_dict()
    method = params["method"]
    del params["method"]
    return {"method": method, "params": [params]}


def json_to_response(json: Dict[str, Any]) -> Response:
    """Converts a JSON response from the rippled server into a Response object.

    Args:
        json: A dictionary representing the contents of the json response from the
              rippled server.

    Returns:
        A Response object containing the information in the rippled server's response.
    """
    result = json["result"]
    raw_status = result["status"]
    if raw_status == "success":
        status = ResponseStatus.SUCCESS
    else:
        status = ResponseStatus.ERROR
    del result["status"]
    # TODO: response_type changes based on what we're getting back... where/how do we
    #  differentiate based on that?
    # TODO: should we pull fields "status" OUT of result dict?
    response_type = ResponseType.RESPONSE
    return Response(status=status, result=result, type=response_type)


def request_to_websocket(request_object: Request) -> Dict[str, Any]:
    """Converts a request model object to the appropriate format for
    interacting with the rippled API via the WebSocket API.

    Args:
        request_object: A Request object representing the parameters of a
                        request to the rippled JSON RPC.

    Returns:
        A dictionary containing the attributes of this Request object formatted
        for submission to the rippled WebSocket API.
    """
    dictionary = request_object.to_dict()
    method = dictionary["method"]
    del dictionary["method"]
    return {**dictionary, "command": method}


def websocket_to_response(response_dict: Dict[str, Any]) -> Response:
    """Converts a WebSocket API response from the rippled server into a Response object.

    Args:
        response_dict: A dictionary representing the contents of the WebSocket API
                       response from the rippled server.

    Returns:
        A Response object containing the information in the rippled server's response.
    """
    raw_status = response_dict["status"]
    request_id = response_dict["id"]
    if raw_status == "success":
        status = ResponseStatus.SUCCESS
        result = response_dict["result"]
    else:
        status = ResponseStatus.ERROR
        result = response_dict
    # TODO: response_type changes based on what we're getting back... where/how do we
    #  differentiate based on that?
    # TODO: should we pull fields "status" OUT of result dict?
    response_type = ResponseType.RESPONSE
    return Response(status=status, result=result, id=request_id, type=response_type)



"""
The ledger_current method returns the unique
identifiers of the current in-progress ledger.
This command is mostly useful for testing,
because the ledger returned is still in flux.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class LedgerCurrent(Request):
    """
    The ledger_current method returns the unique
    identifiers of the current in-progress ledger.
    This command is mostly useful for testing,
    because the ledger returned is still in flux.
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER_CURRENT, init=False)



"""Model for AMMVote transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.amm_create import AMM_MAX_TRADING_FEE
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMVote(Transaction):
    """
    AMMVote is used for submitting a vote for the trading fee of an AMM Instance.

    Any XRPL account that holds LPToken for an AMM instance may submit this
    transaction to vote for the trading fee for that instance.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance.
    """

    trading_fee: int = REQUIRED  # type: ignore
    """
    Specifies the fee, in basis point.
    Valid values for this field are between 0 and 1000 inclusive.
    A value of 1 is equivalent to 1/10 bps or 0.001%, allowing trading fee
    between 0% and 1%. This field is required.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_VOTE,
        init=False,
    )

    def _get_errors(self: AMMVote) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "trading_fee": self._get_trading_fee_error(),
            }.items()
            if value is not None
        }

    def _get_trading_fee_error(self: AMMVote) -> Optional[str]:
        if self.trading_fee < 0 or self.trading_fee > AMM_MAX_TRADING_FEE:
            return f"Must be between 0 and {AMM_MAX_TRADING_FEE}"
        return None



"""Model for AMMBid transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

from typing_extensions import Final

from xrpl.models.amounts import Amount
from xrpl.models.auth_account import AuthAccount
from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

_MAX_AUTH_ACCOUNTS: Final[int] = 4


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMBid(Transaction):
    """
    AMMBid is used to place a bid for the auction slot of obtaining trading advantages
    of an AMM instance.

    An AMM instance auctions off the trading advantages to users (arbitrageurs) at a
    discounted TradingFee for a 24 hour slot.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance.
    """

    bid_min: Optional[Amount] = None
    """
    This field represents the minimum price that the bidder wants to pay for the slot.
    It is specified in units of LPToken. If specified let MinSlotPrice be X and let
    the slot-price computed by price scheduling algorithm be Y, then bidder always pays
    the max(X, Y).
    """

    bid_max: Optional[Amount] = None
    """
    This field represents the maximum price that the bidder wants to pay for the slot.
    It is specified in units of LPToken.
    """

    auth_accounts: Optional[List[AuthAccount]] = None
    """
    This field represents an array of XRPL account IDs that are authorized to trade
    at the discounted fee against the AMM instance.
    A maximum of four accounts can be provided.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_BID,
        init=False,
    )

    def _get_errors(self: AMMBid) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "auth_accounts": self._get_auth_accounts_error(),
            }.items()
            if value is not None
        }

    def _get_auth_accounts_error(self: AMMBid) -> Optional[str]:
        if (
            self.auth_accounts is not None
            and len(self.auth_accounts) > _MAX_AUTH_ACCOUNTS
        ):
            return f"Length must not be greater than {_MAX_AUTH_ACCOUNTS}"
        return None



"""
The tx method retrieves information on a single transaction.

`See tx <https://xrpl.org/tx.html>`_
"""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Tx(Request):
    """
    The tx method retrieves information on a single transaction.

    `See tx <https://xrpl.org/tx.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.TX, init=False)
    transaction: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    binary: bool = False
    min_ledger: Optional[int] = None
    max_ledger: Optional[int] = None



"""This request retrieves information about an AMM instance."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from xrpl.models.currencies import Currency
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMInfo(Request):
    """
    This request retrieves information about an AMM instance.

    Must provide Asset and Asset2 params.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance.
    """

    method: RequestMethod = field(default=RequestMethod.AMM_INFO, init=False)

    def _get_errors(self: AMMInfo) -> Dict[str, str]:
        errors = super()._get_errors()
        return errors



"""Functions for encoding and decoding XRP Ledger addresses and seeds."""
from xrpl.core.addresscodec.codec import (
    SEED_LENGTH,
    decode_account_public_key,
    decode_classic_address,
    decode_node_public_key,
    decode_seed,
    encode_account_public_key,
    encode_classic_address,
    encode_node_public_key,
    encode_seed,
    is_valid_classic_address,
)
from xrpl.core.addresscodec.exceptions import XRPLAddressCodecException
from xrpl.core.addresscodec.main import (
    classic_address_to_xaddress,
    is_valid_xaddress,
    xaddress_to_classic_address,
)
from xrpl.core.addresscodec.utils import XRPL_ALPHABET

__all__ = [
    "classic_address_to_xaddress",
    "decode_account_public_key",
    "decode_classic_address",
    "decode_node_public_key",
    "decode_seed",
    "encode_seed",
    "encode_account_public_key",
    "encode_classic_address",
    "encode_node_public_key",
    "is_valid_classic_address",
    "is_valid_xaddress",
    "SEED_LENGTH",
    "xaddress_to_classic_address",
    "XRPLAddressCodecException",
    "XRPL_ALPHABET",
]



"""
The XRP Ledger has two kinds of money: XRP, and issued
currencies. Both types have high precision, although their
formats are different.
"""
from typing import Union

from xrpl.models.currencies.issued_currency import IssuedCurrency
from xrpl.models.currencies.xrp import XRP

Currency = Union[IssuedCurrency, XRP]



"""Model for EscrowFinish transaction type."""
from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class EscrowFinish(Transaction):
    """
    Represents an `EscrowFinish <https://xrpl.org/escrowfinish.html>`_
    transaction, delivers XRP from a held payment to the recipient.
    """

    owner: str = REQUIRED  # type: ignore
    """
    The source account that funded the Escrow. This field is required.

    :meta hide-value:
    """

    offer_sequence: int = REQUIRED  # type: ignore
    """
    Transaction sequence (or Ticket number) of the EscrowCreate transaction
    that created the Escrow. This field is required.

    :meta hide-value:
    """

    condition: Optional[str] = None
    """
    The previously-supplied `PREIMAGE-SHA-256 crypto-condition
    <https://tools.ietf.org/html/draft-thomas-crypto-conditions-04#section-8.1.>`_
    of the Escrow, if any, as hexadecimal.
    """

    fulfillment: Optional[str] = None
    """
    The `PREIMAGE-SHA-256 crypto-condition fulfillment
    <https://tools.ietf.org/html/draft-thomas-crypto-conditions-04#section-8.1.4.>`_
    matching the Escrow's condition, if any, as hexadecimal.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ESCROW_FINISH,
        init=False,
    )

    def _get_errors(self: EscrowFinish) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.condition and not self.fulfillment:
            errors[
                "fulfillment"
            ] = "If condition is specified, fulfillment must also be specified."
        if self.fulfillment and not self.condition:
            errors[
                "condition"
            ] = "If fulfillment is specified, condition must also be specified."

        return errors



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xrpl',
 'xrpl.account',
 'xrpl.asyncio',
 'xrpl.asyncio.account',
 'xrpl.asyncio.clients',
 'xrpl.asyncio.ledger',
 'xrpl.asyncio.transaction',
 'xrpl.asyncio.wallet',
 'xrpl.clients',
 'xrpl.core',
 'xrpl.core.addresscodec',
 'xrpl.core.binarycodec',
 'xrpl.core.binarycodec.binary_wrappers',
 'xrpl.core.binarycodec.definitions',
 'xrpl.core.binarycodec.types',
 'xrpl.core.keypairs',
 'xrpl.ledger',
 'xrpl.models',
 'xrpl.models.amounts',
 'xrpl.models.currencies',
 'xrpl.models.requests',
 'xrpl.models.transactions',
 'xrpl.models.transactions.pseudo_transactions',
 'xrpl.models.transactions.types',
 'xrpl.transaction',
 'xrpl.utils',
 'xrpl.utils.txn_parser',
 'xrpl.utils.txn_parser.utils',
 'xrpl.wallet']

package_data = \
{'': ['*']}

install_requires = \
['Deprecated>=1.2.13,<2.0.0',
 'ECPy>=1.2.5,<2.0.0',
 'base58>=2.1.0,<3.0.0',
 'httpx>=0.18.1,<0.24.0',
 'pycryptodome>=3.16.0,<4.0.0',
 'types-Deprecated>=1.2.9,<2.0.0',
 'typing-extensions>=4.2.0,<5.0.0']

extras_require = \
{':python_version >= "3.10" and python_version < "4.0"': ['websockets>=10.0,<11.0'],
 ':python_version >= "3.7" and python_version < "3.10"': ['websockets>=9.0.1,<11.0']}

setup_kwargs = {
    'name': 'xrpl-py',
    'version': '1.8.0b2',
    'description': 'A complete Python library for interacting with the XRP ledger',
    'long_description': '[![Documentation Status](https://readthedocs.org/projects/xrpl-py/badge)](https://xrpl-py.readthedocs.io/)\n\n# xrpl-py\n\nA pure Python implementation for interacting with the XRP Ledger, the `xrpl-py` library simplifies the hardest parts of XRP Ledger interaction, like serialization and transaction signing, by providing native Python methods and models for [XRP Ledger transactions](https://xrpl.org/transaction-formats.html) and core server [API](https://xrpl.org/api-conventions.html) ([`rippled`](https://github.com/ripple/rippled)) objects.\n\n\n\n```py\n# create a network client\nfrom xrpl.clients import JsonRpcClient\nclient = JsonRpcClient("https://s.altnet.rippletest.net:51234")\n\n# create a wallet on the testnet\nfrom xrpl.wallet import generate_faucet_wallet\ntest_wallet = generate_faucet_wallet(client)\nprint(test_wallet)\npublic_key: ED3CC1BBD0952A60088E89FA502921895FC81FBD79CAE9109A8FE2D23659AD5D56\nprivate_key: -HIDDEN-\nclassic_address: rBtXmAdEYcno9LWRnAGfT9qBxCeDvuVRZo\n\n# look up account info\nfrom xrpl.models.requests.account_info import AccountInfo\nacct_info = AccountInfo(\n    account="rBtXmAdEYcno9LWRnAGfT9qBxCeDvuVRZo",\n    ledger_index="current",\n    queue=True,\n    strict=True,\n)\nresponse = client.request(acct_info)\nresult = response.result\nimport json\nprint(json.dumps(result["account_data"], indent=4, sort_keys=True))\n# {\n#     "Account": "rBtXmAdEYcno9LWRnAGfT9qBxCeDvuVRZo",\n#     "Balance": "1000000000",\n#     "Flags": 0,\n#     "LedgerEntryType": "AccountRoot",\n#     "OwnerCount": 0,\n#     "PreviousTxnID": "73CD4A37537A992270AAC8472F6681F44E400CBDE04EC8983C34B519F56AB107",\n#     "PreviousTxnLgrSeq": 16233962,\n#     "Sequence": 16233962,\n#     "index": "FD66EC588B52712DCE74831DCB08B24157DC3198C29A0116AA64D310A58512D7"\n# }\n```\n\n[![Downloads](https://pepy.tech/badge/xrpl-py/month)](https://pepy.tech/project/xrpl-py/month)\n[![Contributors](https://img.shields.io/github/contributors/xpring-eng/xrpl-py.svg)](https://github.com/xpring-eng/xrpl-py/graphs/contributors)\n\n## Installation and supported versions\n\nThe `xrpl-py` library is available on [PyPI](https://pypi.org/). Install with `pip`:\n\n\n```\npip3 install xrpl-py\n```\n\nThe library supports [Python 3.7](https://www.python.org/downloads/) and later.\n\n[![Supported Versions](https://img.shields.io/pypi/pyversions/xrpl-py.svg)](https://pypi.org/project/xrpl-py)\n\n\n## Features\n\nUse `xrpl-py` to build Python applications that leverage the [XRP Ledger](https://xrpl.org/). The library helps with all aspects of interacting with the XRP Ledger, including:\n\n* Key and wallet management\n* Serialization\n* Transaction Signing\n\n`xrpl-py` also provides:\n\n* A network client — See [`xrpl.clients`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.clients.html) for more information.\n* Methods for inspecting accounts — See [XRPL Account Methods](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.account.html) for more information.\n* Codecs for encoding and decoding addresses and other objects — See [Core Codecs](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.core.html) for more information.\n\n## [➡️ Reference Documentation](https://xrpl-py.readthedocs.io/en/stable/)\n\nSee the complete [`xrpl-py` reference documentation on Read the Docs](https://xrpl-py.readthedocs.io/en/stable/index.html).\n\n\n## Usage\n\nThe following sections describe some of the most commonly used modules in the `xrpl-py` library and provide sample code.\n\n### Network client\n\nUse the `xrpl.clients` library to create a network client for connecting to the XRP Ledger.\n\n```py\nfrom xrpl.clients import JsonRpcClient\nJSON_RPC_URL = "https://s.altnet.rippletest.net:51234"\nclient = JsonRpcClient(JSON_RPC_URL)\n```\n\n### Manage keys and wallets\n\n#### `xrpl.wallet`\n\nUse the [`xrpl.wallet`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.wallet.html) module to create a wallet from a given seed or or via a [Testnet faucet](https://xrpl.org/xrp-testnet-faucet.html).\n\nTo create a wallet from a seed (in this case, the value generated using [`xrpl.keypairs`](#xrpl-keypairs)):\n\n```py\nwallet_from_seed = xrpl.wallet.Wallet(seed, 0)\nprint(wallet_from_seed)\n# pub_key: ED46949E414A3D6D758D347BAEC9340DC78F7397FEE893132AAF5D56E4D7DE77B0\n# priv_key: -HIDDEN-\n# classic_address: rG5ZvYsK5BPi9f1Nb8mhFGDTNMJhEhufn6\n```\n\nTo create a wallet from a Testnet faucet:\n\n```py\ntest_wallet = generate_faucet_wallet(client)\ntest_account = test_wallet.classic_address\nprint("Classic address:", test_account)\n# Classic address: rEQB2hhp3rg7sHj6L8YyR4GG47Cb7pfcuw\n```\n\n#### `xrpl.core.keypairs`\n\nUse the [`xrpl.core.keypairs`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.core.keypairs.html#module-xrpl.core.keypairs) module to generate seeds and derive keypairs and addresses from those seed values.\n\nHere\'s an example of how to generate a `seed` value and derive an [XRP Ledger "classic" address](https://xrpl.org/cryptographic-keys.html#account-id-and-address) from that seed.\n\n\n```py\nfrom xrpl.core import keypairs\nseed = keypairs.generate_seed()\npublic, private = keypairs.derive_keypair(seed)\ntest_account = keypairs.derive_classic_address(public)\nprint("Here\'s the public key:")\nprint(public)\nprint("Here\'s the private key:")\nprint(private)\nprint("Store this in a secure place!")\n# Here\'s the public key:\n# ED3CC1BBD0952A60088E89FA502921895FC81FBD79CAE9109A8FE2D23659AD5D56\n# Here\'s the private key:\n# EDE65EE7882847EF5345A43BFB8E6F5EEC60F45461696C384639B99B26AAA7A5CD\n# Store this in a secure place!\n```\n\n**Note:** You can use `xrpl.core.keypairs.sign` to sign transactions but `xrpl-py` also provides explicit methods for safely signing and submitting transactions. See [Transaction Signing](#transaction-signing) and [XRPL Transaction Methods](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.transaction.html#module-xrpl.transaction) for more information.\n\n\n### Serialize and sign transactions\n\nTo securely submit transactions to the XRP Ledger, you need to first serialize data from JSON and other formats into the [XRP Ledger\'s canonical format](https://xrpl.org/serialization.html), then to [authorize the transaction](https://xrpl.org/transaction-basics.html#authorizing-transactions) by digitally [signing it](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.core.keypairs.html?highlight=sign#xrpl.core.keypairs.sign) with the account\'s private key. The `xrpl-py` library provides several methods to simplify this process.\n\n\nUse the [`xrpl.transaction`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.transaction.html) module to sign and submit transactions. The module offers three ways to do this:\n\n* [`sign_and_submit`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.transaction.html#xrpl.transaction.sign_and_submit) — Signs a transaction locally, then submits it to the XRP Ledger. This method does not implement [reliable transaction submission](https://xrpl.org/reliable-transaction-submission.html#reliable-transaction-submission) best practices, so only use it for development or testing purposes.\n\n* [`sign`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.transaction.html#xrpl.transaction.sign) — Signs a transaction locally. This method **does  not** submit the transaction to the XRP Ledger.\n\n* [`send_reliable_submission`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.transaction.html#xrpl.transaction.send_reliable_submission) — An implementation of the [reliable transaction submission guidelines](https://xrpl.org/reliable-transaction-submission.html#reliable-transaction-submission), this method submits a signed transaction to the XRP Ledger and then verifies that it has been included in a validated ledger (or has failed to do so). Use this method to submit transactions for production purposes.\n\n\n```py\nfrom xrpl.models.transactions import Payment\nfrom xrpl.transaction import sign, send_reliable_submission\nfrom xrpl.ledger import get_latest_validated_ledger_sequence\nfrom xrpl.account import get_next_valid_seq_number\n\ncurrent_validated_ledger = get_latest_validated_ledger_sequence(client)\ntest_wallet.sequence = get_next_valid_seq_number(test_wallet.classic_address, client)\n\n# prepare the transaction\n# the amount is expressed in drops, not XRP\n# see https://xrpl.org/basic-data-types.html#specifying-currency-amounts\nmy_tx_payment = Payment(\n    account=test_wallet.classic_address,\n    amount="2200000",\n    destination="rPT1Sjq2YGrBMTttX4GZHjKu9dyfzbpAYe",\n    last_ledger_sequence=current_validated_ledger + 20,\n    sequence=test_wallet.sequence,\n    fee="10",\n)\n# sign the transaction\nmy_tx_payment_signed = sign(my_tx_payment,test_wallet)\n\n# submit the transaction\ntx_response = send_reliable_submission(my_tx_payment_signed, client)\n```\n\n#### Get fee from the XRP Ledger\n\n\nIn most cases, you can specify the minimum [transaction cost](https://xrpl.org/transaction-cost.html#current-transaction-cost) of `"10"` for the `fee` field unless you have a strong reason not to. But if you want to get the [current load-balanced transaction cost](https://xrpl.org/transaction-cost.html#current-transaction-cost) from the network, you can use the `get_fee` function:\n\n```py\nfrom xrpl.ledger import get_fee\nfee = get_fee(client)\nprint(fee)\n# 10\n```\n\n#### Auto-filled fields\n\nThe `xrpl-py` library automatically populates the `fee`, `sequence` and `last_ledger_sequence` fields when you create transactions. In the example above, you could omit those fields and let the library fill them in for you.\n\n```py\nfrom xrpl.models.transactions import Payment\nfrom xrpl.transaction import send_reliable_submission, autofill_and_sign\n# prepare the transaction\n# the amount is expressed in drops, not XRP\n# see https://xrpl.org/basic-data-types.html#specifying-currency-amounts\nmy_tx_payment = Payment(\n    account=test_wallet.classic_address,\n    amount="2200000",\n    destination="rPT1Sjq2YGrBMTttX4GZHjKu9dyfzbpAYe"\n)\n\n# sign the transaction with the autofill method\n# (this will auto-populate the fee, sequence, and last_ledger_sequence)\nmy_tx_payment_signed = autofill_and_sign(my_tx_payment, test_wallet, client)\nprint(my_tx_payment_signed)\n# Payment(\n#     account=\'rMPUKmzmDWEX1tQhzQ8oGFNfAEhnWNFwz\',\n#     transaction_type=<TransactionType.PAYMENT: \'Payment\'>,\n#     fee=\'10\',\n#     sequence=16034065,\n#     account_txn_id=None,\n#     flags=0,\n#     last_ledger_sequence=10268600,\n#     memos=None,\n#     signers=None,\n#     source_tag=None,\n#     signing_pub_key=\'EDD9540FA398915F0BCBD6E65579C03BE5424836CB68B7EB1D6573F2382156B444\',\n#     txn_signature=\'938FB22AE7FE76CF26FD11F8F97668E175DFAABD2977BCA397233117E7E1C4A1E39681091CC4D6DF21403682803AB54CC21DC4FA2F6848811DEE10FFEF74D809\',\n#     amount=\'2200000\',\n#     destination=\'rPT1Sjq2YGrBMTttX4GZHjKu9dyfzbpAYe\',\n#     destination_tag=None,\n#     invoice_id=None,\n#     paths=None,\n#     send_max=None,\n#     deliver_min=None\n# )\n\n# submit the transaction\ntx_response = send_reliable_submission(my_tx_payment_signed, client)\n```\n\n\n### Subscribe to ledger updates\n\nYou can send `subscribe` and `unsubscribe` requests only using the WebSocket network client. These request methods allow you to be alerted of certain situations as they occur, such as when a new ledger is declared.\n\n```py\nfrom xrpl.clients import WebsocketClient\nurl = "wss://s.altnet.rippletest.net/"\nfrom xrpl.models.requests import Subscribe, StreamParameter\nreq = Subscribe(streams=[StreamParameter.LEDGER])\n# NOTE: this code will run forever without a timeout, until the process is killed\nwith WebsocketClient(url) as client:\n    client.send(req)\n    for message in client:\n        print(message)\n# {\'result\': {\'fee_base\': 10, \'fee_ref\': 10, \'ledger_hash\': \'7CD50477F23FF158B430772D8E82A961376A7B40E13C695AA849811EDF66C5C0\', \'ledger_index\': 18183504, \'ledger_time\': 676412962, \'reserve_base\': 20000000, \'reserve_inc\': 5000000, \'validated_ledgers\': \'17469391-18183504\'}, \'status\': \'success\', \'type\': \'response\'}\n# {\'fee_base\': 10, \'fee_ref\': 10, \'ledger_hash\': \'BAA743DABD168BD434804416C8087B7BDEF7E6D7EAD412B9102281DD83B10D00\', \'ledger_index\': 18183505, \'ledger_time\': 676412970, \'reserve_base\': 20000000, \'reserve_inc\': 5000000, \'txn_count\': 0, \'type\': \'ledgerClosed\', \'validated_ledgers\': \'17469391-18183505\'}\n# {\'fee_base\': 10, \'fee_ref\': 10, \'ledger_hash\': \'D8227DAF8F745AE3F907B251D40B4081E019D013ABC23B68C0B1431DBADA1A46\', \'ledger_index\': 18183506, \'ledger_time\': 676412971, \'reserve_base\': 20000000, \'reserve_inc\': 5000000, \'txn_count\': 0, \'type\': \'ledgerClosed\', \'validated_ledgers\': \'17469391-18183506\'}\n# {\'fee_base\': 10, \'fee_ref\': 10, \'ledger_hash\': \'CFC412B6DDB9A402662832A781C23F0F2E842EAE6CFC539FEEB287318092C0DE\', \'ledger_index\': 18183507, \'ledger_time\': 676412972, \'reserve_base\': 20000000, \'reserve_inc\': 5000000, \'txn_count\': 0, \'type\': \'ledgerClosed\', \'validated_ledgers\': \'17469391-18183507\'}\n```\n\n\n### Asynchronous Code\n\nThis library supports Python\'s [`asyncio`](https://docs.python.org/3/library/asyncio.html) package, which is used to run asynchronous code. All the async code is in [`xrpl.asyncio`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.asyncio.html) If you are writing asynchronous code, please note that you will not be able to use any synchronous sugar functions, due to how event loops are handled. However, every synchronous method has a corresponding asynchronous method that you can use.\n\nThis sample code is the asynchronous equivalent of the above section on submitting a transaction.\n\n```py\nimport asyncio\nfrom xrpl.models.transactions import Payment\nfrom xrpl.asyncio.transaction import sign, send_reliable_submission\nfrom xrpl.asyncio.ledger import get_latest_validated_ledger_sequence\nfrom xrpl.asyncio.account import get_next_valid_seq_number\nfrom xrpl.asyncio.clients import AsyncJsonRpcClient\n\nasync_client = AsyncJsonRpcClient(JSON_RPC_URL)\n\nasync def submit_sample_transaction():\n    current_validated_ledger = await get_latest_validated_ledger_sequence(async_client)\n    test_wallet.sequence = await get_next_valid_seq_number(test_wallet.classic_address, async_client)\n\n    # prepare the transaction\n    # the amount is expressed in drops, not XRP\n    # see https://xrpl.org/basic-data-types.html#specifying-currency-amounts\n    my_tx_payment = Payment(\n        account=test_wallet.classic_address,\n        amount="2200000",\n        destination="rPT1Sjq2YGrBMTttX4GZHjKu9dyfzbpAYe",\n        last_ledger_sequence=current_validated_ledger + 20,\n        sequence=test_wallet.sequence,\n        fee="10",\n    )\n    # sign the transaction\n    my_tx_payment_signed = await sign(my_tx_payment,test_wallet)\n\n    # submit the transaction\n    tx_response = await send_reliable_submission(my_tx_payment_signed, async_client)\n\nasyncio.run(submit_sample_transaction())\n```\n\n### Encode addresses\n\nUse [`xrpl.core.addresscodec`](https://xrpl-py.readthedocs.io/en/stable/source/xrpl.core.addresscodec.html) to encode and decode addresses into and from the ["classic" and X-address formats](https://xrpl.org/accounts.html#addresses).\n\n```py\n# convert classic address to x-address\nfrom xrpl.core import addresscodec\ntestnet_xaddress = (\n    addresscodec.classic_address_to_xaddress(\n        "rMPUKmzmDWEX1tQhzQ8oGFNfAEhnWNFwz",\n        tag=0,\n        is_test_network=True,\n    )\n)\nprint(testnet_xaddress)\n# T7QDemmxnuN7a52A62nx2fxGPWcRahLCf3qaswfrsNW9Lps\n```\n\n\n## Contributing\n\nIf you want to contribute to this project, see [CONTRIBUTING.md].\n\n### Mailing Lists\n\nWe have a low-traffic mailing list for announcements of new `xrpl-py` releases. (About 1 email per week)\n\n+ [Subscribe to xrpl-announce](https://groups.google.com/g/xrpl-announce)\n\nIf you\'re using the XRP Ledger in production, you should run a [rippled server](https://github.com/ripple/rippled) and subscribe to the ripple-server mailing list as well.\n\n+ [Subscribe to ripple-server](https://groups.google.com/g/ripple-server)\n\n### Code Samples\n- For samples of common use cases, see the [XRPL.org Code Samples](https://xrpl.org/code-samples.html) page.\n- You can also browse those samples [directly on GitHub](https://github.com/XRPLF/xrpl-dev-portal/tree/master/content/_code-samples).\n\n### Report an issue\n\nExperienced an issue? Report it [here](https://github.com/XRPLF/xrpl-py/issues/new).\n\n## License\n\nThe `xrpl-py` library is licensed under the ISC License. See [LICENSE] for more information.\n\n\n\n[CONTRIBUTING.md]: CONTRIBUTING.md\n[LICENSE]: LICENSE\n',
    'author': 'Mayukha Vadari',
    'author_email': 'mvadari@ripple.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/XRPLF/xrpl-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Codec for currency property inside an XRPL issued currency amount json."""
from __future__ import annotations  # Requires Python 3.7+

from typing import Optional, Type

from typing_extensions import Final

from xrpl.constants import HEX_CURRENCY_REGEX, ISO_CURRENCY_REGEX
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.hash160 import Hash160

_CURRENCY_CODE_LENGTH: Final[int] = 20  # bytes


def _is_iso_code(value: str) -> bool:
    """Tests if value is a valid 3-char iso code."""
    return bool(ISO_CURRENCY_REGEX.fullmatch(value))


def _iso_code_from_hex(value: bytes) -> Optional[str]:
    candidate_iso = value.decode("ascii")
    if candidate_iso == "XRP":
        raise XRPLBinaryCodecException(
            "Disallowed currency code: to indicate the currency "
            "XRP you must use 20 bytes of 0s"
        )
    if _is_iso_code(candidate_iso):
        return candidate_iso
    return None


def _is_hex(value: str) -> bool:
    """Tests if value is a valid 40-char hex string."""
    return bool(HEX_CURRENCY_REGEX.fullmatch(value))


def _iso_to_bytes(iso: str) -> bytes:
    """
    Convert an ISO code to a 160-bit (20 byte) encoded representation.

    See "Currency codes" subheading in
    `Amount Fields <https://xrpl.org/serialization.html#amount-fields>`_
    """
    if not _is_iso_code(iso):
        raise XRPLBinaryCodecException(f"Invalid ISO code: {iso}")

    if iso == "XRP":
        # This code (160 bit all zeroes) is used to indicate XRP in
        # rare cases where a field must specify a currency code for XRP.
        return bytes(_CURRENCY_CODE_LENGTH)

    iso_bytes = iso.encode("ASCII")
    # Currency Codes: https://xrpl.org/currency-formats.html#standard-currency-codes
    # 160 total bits:
    #   8 bits type code (0x00)
    #   88 bits reserved (0's)
    #   24 bits ASCII
    #   16 bits version (0x00)
    #   24 bits reserved (0's)
    return bytes(12) + iso_bytes + bytes(5)


class Currency(Hash160):
    """
    Codec for serializing and deserializing currency codes in issued currency amounts.
    `Amount fields <https://xrpl.org/serialization.html#amount-fields>`_

    Attributes:
        buffer: The byte encoding of this currency.
        _iso: The three-character ISO currency code if standard format, else None.
    """

    LENGTH: Final[int] = 20
    _iso: Optional[str] = None

    def __init__(self: Currency, buffer: Optional[bytes] = None) -> None:
        """Construct a Currency."""
        if buffer is not None:
            super().__init__(buffer)
        else:
            super().__init__(bytes(self.LENGTH))

        code_bytes = self.buffer[12:15]
        # Determine whether this currency code is in standard or nonstandard format:
        # https://xrpl.org/currency-formats.html#nonstandard-currency-codes
        if self.buffer[0] != 0:
            # non-standard currency
            self._iso = None
        elif self.buffer.hex() == "0" * 40:  # all 0s
            # the special case for literal XRP
            self._iso = "XRP"
        else:
            self._iso = _iso_code_from_hex(code_bytes)

    @classmethod
    def from_value(cls: Type[Currency], value: str) -> Currency:
        """
        Construct a Currency object from a string representation of a currency.

        Args:
            value: The string to construct a Currency object from.

        Returns:
            A Currency object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the Currency representation is invalid.
        """
        if not isinstance(value, str):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a Currency: expected str,"
                f" received {value.__class__.__name__}."
            )

        if _is_iso_code(value):
            return Currency(_iso_to_bytes(value))
        if _is_hex(value):
            return cls(bytes.fromhex(value))
        raise XRPLBinaryCodecException("Unsupported Currency representation: {value}")

    def to_json(self: Currency) -> str:
        """
        Returns the JSON representation of a currency.

        Returns:
            The JSON representation of a Currency.
        """
        if self._iso is not None:
            return self._iso
        return self.buffer.hex().upper()



"""Model for a XChainAddClaimAttestation transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from typing_extensions import Literal

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainAddClaimAttestation(Transaction):
    """Represents a XChainAddClaimAttestation transaction."""

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore

    public_key: str = REQUIRED  # type: ignore

    signature: str = REQUIRED  # type: ignore

    other_chain_source: str = REQUIRED  # type: ignore

    amount: Amount = REQUIRED  # type: ignore

    attestation_reward_account: str = REQUIRED  # type: ignore

    attestation_signer_account: str = REQUIRED  # type: ignore

    was_locking_chain_send: Union[Literal[0], Literal[1]] = REQUIRED  # type: ignore

    xchain_claim_id: str = REQUIRED  # type: ignore

    destination: Optional[str] = None

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_ADD_CLAIM_ATTESTATION,
        init=False,
    )



"""Model for NFTokenCreateOffer transaction type and related flag."""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from xrpl.models.amounts import Amount, get_amount_value
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class NFTokenCreateOfferFlag(int, Enum):
    """Transaction Flags for an NFTokenCreateOffer Transaction."""

    TF_SELL_NFTOKEN = 0x00000001
    """
    If set, indicates that the offer is a sell offer.
    Otherwise, it is a buy offer.
    """


class NFTokenCreateOfferFlagInterface(FlagInterface):
    """Transaction Flags for an NFTokenCreateOffer Transaction."""

    TF_SELL_NFTOKEN: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class NFTokenCreateOffer(Transaction):
    """
    The NFTokenCreateOffer transaction creates either an offer to buy an
    NFT the submitting account does not own, or an offer to sell an NFT
    the submitting account does own.
    """

    nftoken_id: str = REQUIRED  # type: ignore
    """
    Identifies the TokenID of the NFToken object that the
    offer references. This field is required.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    Indicates the amount expected or offered for the Token.

    The amount must be non-zero, except when this is a sell
    offer and the asset is XRP. This would indicate that the current
    owner of the token is giving it away free, either to anyone at all,
    or to the account identified by the Destination field. This field
    is required.

    :meta hide-value:
    """

    owner: Optional[str] = None
    """
    Indicates the AccountID of the account that owns the
    corresponding NFToken.

    If the offer is to buy a token, this field must be present
    and it must be different than Account (since an offer to
    buy a token one already holds is meaningless).

    If the offer is to sell a token, this field must not be
    present, as the owner is, implicitly, the same as Account
    (since an offer to sell a token one doesn't already hold
    is meaningless).
    """

    expiration: Optional[int] = None
    """
    Indicates the time after which the offer will no longer
    be valid. The value is the number of seconds since the
    Ripple Epoch.
    """

    destination: Optional[str] = None
    """
    If present, indicates that this offer may only be
    accepted by the specified account. Attempts by other
    accounts to accept this offer MUST fail.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_CREATE_OFFER,
        init=False,
    )

    def _get_errors(self: NFTokenCreateOffer) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "amount": self._get_amount_error(),
                "destination": self._get_destination_error(),
                "owner": self._get_owner_error(),
            }.items()
            if value is not None
        }

    def _get_amount_error(self: NFTokenCreateOffer) -> Optional[str]:
        if (
            not self.has_flag(NFTokenCreateOfferFlag.TF_SELL_NFTOKEN)
            and get_amount_value(self.amount) <= 0
        ):
            return "Must be greater than 0 for a buy offer"
        return None

    def _get_destination_error(self: NFTokenCreateOffer) -> Optional[str]:
        if self.destination == self.account:
            return "Must not be equal to the account"
        return None

    def _get_owner_error(self: NFTokenCreateOffer) -> Optional[str]:
        if (
            not self.has_flag(NFTokenCreateOfferFlag.TF_SELL_NFTOKEN)
            and self.owner is None
        ):
            return "Must be present for buy offers"
        if (
            self.has_flag(NFTokenCreateOfferFlag.TF_SELL_NFTOKEN)
            and self.owner is not None
        ):
            return "Must not be present for sell offers"
        if self.owner == self.account:
            return "Must not be equal to the account"
        return None



"""High-level methods to obtain information about account transaction history."""
import asyncio
from typing import Any, Dict, List

from deprecated.sphinx import deprecated

from xrpl.asyncio.account import transaction_history
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response


def get_latest_transaction(account: str, client: SyncClient) -> Response:
    """
    Fetches the most recent transaction on the ledger associated with an account.

    Args:
        account: the account to query.
        client: the network client used to communicate with a rippled node.

    Returns:
        The Response object containing the transaction info.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    return asyncio.run(transaction_history.get_latest_transaction(account, client))


@deprecated(
    reason="Sending an AccountTx request directly allows you to page through all "
    "results and is just as easy to use.",
    version="1.6.0",
)
def get_account_transactions(address: str, client: SyncClient) -> List[Dict[str, Any]]:
    """
    Query the ledger for a list of transactions that involved a given account.
    To access more than just the first page of results, use the :class:`AccountTx`
    request directly.

    Args:
        address: the account to query.
        client: the network client used to make network calls.

    Returns:
        The most recent transaction history for the address. For the full history,
        page through the :class:`AccountTx` request directly.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    return asyncio.run(transaction_history.get_account_transactions(address, client))


@deprecated(
    reason="Sending an AccountTx request directly and filtering for payments allows "
    "you to page through all results and is just as easy to use.",
    version="1.8.0",
)
def get_account_payment_transactions(
    address: str, client: SyncClient
) -> List[Dict[str, Any]]:
    """
    Query the ledger for a list of payment transactions that involved a given account.
    To access more than just the first page of results, use the :class:`AccountTx`
    request directly then filter for transactions with a "Payment" TransactionType.

    Args:
        address: the account to query.
        client: the network client used to make network calls.

    Returns:
        The first page of payment transaction history for the address. For the full
        history, page through the :class:`AccountTx` request directly.
    """
    return asyncio.run(
        transaction_history.get_account_payment_transactions(address, client)
    )



"""Interface for cryptographic key pairs for use with the XRP Ledger."""
from secrets import token_bytes
from typing import Dict, Optional, Tuple, Type

from typing_extensions import Final

from xrpl.constants import CryptoAlgorithm
from xrpl.core import addresscodec
from xrpl.core.keypairs.crypto_implementation import CryptoImplementation
from xrpl.core.keypairs.ed25519 import ED25519
from xrpl.core.keypairs.ed25519 import PREFIX as ED_PREFIX
from xrpl.core.keypairs.exceptions import XRPLKeypairsException
from xrpl.core.keypairs.helpers import get_account_id
from xrpl.core.keypairs.secp256k1 import SECP256K1

_VERIFICATION_MESSAGE: Final[bytes] = b"This test message should verify."

_ALGORITHM_TO_MODULE_MAP: Final[Dict[CryptoAlgorithm, Type[CryptoImplementation]]] = {
    CryptoAlgorithm.ED25519: ED25519,
    CryptoAlgorithm.SECP256K1: SECP256K1,
}


def generate_seed(
    entropy: Optional[str] = None,
    algorithm: CryptoAlgorithm = CryptoAlgorithm.ED25519,
) -> str:
    """
    Generate a seed value that cryptographic keys can be derived from.

    Args:
        entropy: Must be at least addresscodec.SEED_LENGTH bytes long and
            will be truncated to that length
        algorithm: CryptoAlgorithm to use for seed generation. The  default is
            :data:`CryptoAlgorithm.ED25519 <xrpl.CryptoAlgorithm.ED25519>`.

    Returns:
        A seed value that can be used to derive a key pair with the given
            cryptographic algorithm.
    """
    if entropy is None:
        parsed_entropy = token_bytes(addresscodec.SEED_LENGTH)
    else:
        parsed_entropy = bytes(entropy, "UTF-8")[: addresscodec.SEED_LENGTH]
    return addresscodec.encode_seed(parsed_entropy, algorithm)


def derive_keypair(
    seed: str, validator: bool = False, algorithm: Optional[CryptoAlgorithm] = None
) -> Tuple[str, str]:
    """
    Derive the public and private keys from a given seed value.

    Args:
        seed: Seed to derive the key pair from. Use
            :func:`generate_seed() <xrpl.core.keypairs.generate_seed>` to generate an
            appropriate value.
        validator: Whether the keypair is a validator keypair.
        algorithm: The algorithm used to encode the keys. Inferred from the seed if not
            included.

    Returns:
        A (public key, private key) pair derived from the given seed.

    Raises:
        XRPLKeypairsException: If the derived keypair did not generate a
            verifiable signature.
    """
    decoded_seed, algorithm = addresscodec.decode_seed(seed, algorithm)
    module = _ALGORITHM_TO_MODULE_MAP[algorithm]
    public_key, private_key = module.derive_keypair(decoded_seed, validator)
    signature = module.sign(_VERIFICATION_MESSAGE, private_key)
    if not module.is_valid_message(_VERIFICATION_MESSAGE, signature, public_key):
        raise XRPLKeypairsException(
            "Derived keypair did not generate verifiable signature",
        )
    return public_key, private_key


def derive_classic_address(public_key: str) -> str:
    """
    Derive the XRP Ledger classic address for a given public key. See
    `Address Derivation
    <https://xrpl.org/cryptographic-keys.html#account-id-and-address>`_
    for more information.

    Args:
        public_key: The public key to derive the address from, as hexadecimal.

    Returns:
        The classic address corresponding to the given public key.
    """
    account_id = get_account_id(bytes.fromhex(public_key))
    return addresscodec.encode_classic_address(account_id)


def sign(message: bytes, private_key: str) -> str:
    """
    Sign a message using a given private key.

    Args:
        message: The message to sign, as bytes.
        private_key: The private key to use to sign the message.

    Returns:
        Signed message, as hexadecimal.
    """
    return (
        _get_module_from_key(private_key)
        .sign(
            message,
            private_key,
        )
        .hex()
        .upper()
    )


def is_valid_message(message: bytes, signature: bytes, public_key: str) -> bool:
    """
    Verifies the signature on a given message.

    Args:
        message: The message to validate.
        signature: The signature of the message.
        public_key: The public key to use to verify the message and
            signature.

    Returns:
        Whether the message is valid for the given signature and public key.
    """
    return _get_module_from_key(public_key).is_valid_message(
        message,
        signature,
        public_key,
    )


def _get_module_from_key(key: str) -> Type[CryptoImplementation]:
    if key.startswith(ED_PREFIX):
        return ED25519
    return SECP256K1



"""This module contains commonly-used constants."""
from typing_extensions import Final

# The dictionary used for XRPL base58 encodings
XRPL_ALPHABET: Final[
    bytes
] = b"rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz"



"""
Conversions between the XRP Ledger's 'Ripple Epoch' time and native time
data types
"""

from datetime import datetime, timezone
from typing import Union

from typing_extensions import Final

from xrpl.constants import XRPLException

RIPPLE_EPOCH: Final[int] = 946684800
"""The "Ripple Epoch" of 2000-01-01T00:00:00 UTC"""

MAX_XRPL_TIME: Final[int] = 2**32
"""The maximum time that can be expressed on the XRPL"""


def ripple_time_to_datetime(ripple_time: int) -> datetime:
    """
    Convert from XRP Ledger 'Ripple Epoch' time to a UTC `datetime
    <https://docs.python.org/3/library/datetime.html#datetime-objects>`_ object.

    Args:
        ripple_time: Whole seconds since the Ripple Epoch of 2001-01-01T00:00Z

    Returns:
        The equivalent time as a ``datetime`` instance.

    Raises:
        XRPLTimeRangeException: if the given ``ripple_time`` is not valid
    """
    if ripple_time < 0:
        raise XRPLTimeRangeException(f"{ripple_time} is before the Ripple Epoch.")
    if ripple_time > MAX_XRPL_TIME:
        raise XRPLTimeRangeException(
            f"{ripple_time} is larger than any time that can be expressed on"
            f"the XRP Ledger."
        )
    timestamp = ripple_time + RIPPLE_EPOCH
    return datetime.fromtimestamp(timestamp, timezone.utc)


def datetime_to_ripple_time(dt: datetime) -> int:
    """
    Convert from a `datetime
    <https://docs.python.org/3/library/datetime.html#datetime-objects>`_ object
    to an XRP Ledger 'Ripple Epoch' time.

    Args:
        dt: The datetime to convert

    Returns:
        The equivalent time in whole seconds since the Ripple Epoch of
        2001-01-01T00:00Z

    Raises:
        XRPLTimeRangeException: if the time is outside the range that can be
                                represented in Ripple Epoch time
    """
    ripple_time = int(dt.timestamp() - RIPPLE_EPOCH)
    if ripple_time < 0:
        raise XRPLTimeRangeException(f"Datetime {dt} is before the Ripple Epoch")
    if ripple_time > MAX_XRPL_TIME:
        raise XRPLTimeRangeException(
            f"{dt} is later than any time that can be expressed on" f"the XRP Ledger."
        )
    return ripple_time


def ripple_time_to_posix(ripple_time: int) -> int:
    """
    Convert from XRP Ledger 'Ripple Epoch' time to a POSIX-like integer
    timestamp.

    Args:
        ripple_time: A timestamp as the number of whole seconds since the
                     Ripple Epoch of 2001-01-01T00:00Z

    Returns:
        The equivalent time in whole seconds since the UNIX epoch of
        1970-01-01T00:00Z

    Raises:
        XRPLTimeRangeException: if the given ``ripple_time`` is not valid
    """
    if ripple_time < 0:
        raise XRPLTimeRangeException(f"{ripple_time} is before the Ripple Epoch.")
    if ripple_time > MAX_XRPL_TIME:
        raise XRPLTimeRangeException(
            f"{ripple_time} is larger than any time that can be expressed on"
            f"the XRP Ledger."
        )
    timestamp = ripple_time + RIPPLE_EPOCH
    return timestamp


def posix_to_ripple_time(timestamp: Union[int, float]) -> int:
    """
    Convert from a POSIX-like timestamp such as one returned by `time.time()
    <https://docs.python.org/3/library/time.html#time.time>`_ to an XRP Ledger
    'Ripple Epoch' time.

    Args:
        timestamp: An integer or floating-point timestamp such as one returned
                   by ``time.time()``.

    Returns:
        The equivalent time in whole seconds since the Ripple Epoch of
        2001-01-01T00:00Z

    Raises:
        XRPLTimeRangeException: if the time is outside the range that can be
                                represented in Ripple Epoch time


    """
    ripple_time = int(timestamp - RIPPLE_EPOCH)
    if ripple_time < 0:
        raise XRPLTimeRangeException(
            f"Timestamp {timestamp} is before the Ripple Epoch."
        )
    if ripple_time > MAX_XRPL_TIME:
        raise XRPLTimeRangeException(
            f"{timestamp} is later than any time that can be expressed on"
            f"the XRP Ledger."
        )
    return ripple_time


class XRPLTimeRangeException(XRPLException):
    """Exception for invalid XRP Ledger time data."""

    pass



"""Model for a XChainClaim transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional, Union

from xrpl.models.amounts import Amount
from xrpl.models.currencies import XRP
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainClaim(Transaction):
    """Represents a XChainClaim transaction."""

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore

    xchain_claim_id: Union[int, str] = REQUIRED  # type: ignore

    destination: str = REQUIRED  # type: ignore

    destination_tag: Optional[int] = None

    amount: Amount = REQUIRED  # type: ignore

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_CLAIM,
        init=False,
    )

    def _get_errors(self: XChainClaim) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge
        currency = XRP() if isinstance(self.amount, str) else self.amount.to_currency()
        if (
            currency != bridge.locking_chain_issue
            and currency != bridge.issuing_chain_issue
        ):
            errors[
                "amount"
            ] = "Amount must match either locking chain issue or issuing chain issue."

        return errors



"""High-level reliable submission methods with XRPL transactions."""

import asyncio

from typing_extensions import Final

from xrpl.asyncio.clients import Client
from xrpl.asyncio.ledger import get_latest_validated_ledger_sequence
from xrpl.asyncio.transaction.main import submit
from xrpl.clients import XRPLRequestFailureException
from xrpl.constants import XRPLException
from xrpl.models.requests import Tx
from xrpl.models.response import Response
from xrpl.models.transactions.transaction import Transaction

_LEDGER_CLOSE_TIME: Final[int] = 1


class XRPLReliableSubmissionException(XRPLException):
    """General XRPL Reliable Submission Exception."""

    pass


async def _wait_for_final_transaction_outcome(
    transaction_hash: str, client: Client, prelim_result: str, attempts: int = 0
) -> Response:
    """
    The core logic of reliable submission.  Polls the ledger until the result of the
    transaction can be considered final, meaning it has either been included in a
    validated ledger, or the transaction's lastLedgerSequence has been surpassed by the
    latest ledger sequence (meaning it will never be included in a validated ledger).
    """
    await asyncio.sleep(_LEDGER_CLOSE_TIME)
    # new persisted transaction

    # query transaction by hash
    try:
        transaction_response = await client._request_impl(
            Tx(transaction=transaction_hash)
        )
    except XRPLRequestFailureException as e:
        if e.error == "txnNotFound" and attempts < 4:

            """
            For the case if a submitted transaction is still
            in queue and not processed on the ledger yet.
            Retry 4 times before raising an exception.
            """
            return await _wait_for_final_transaction_outcome(
                transaction_hash, client, prelim_result, attempts + 1
            )
        else:
            raise e
    result = transaction_response.result
    if "validated" in result and result["validated"]:
        # result is in a validated ledger, outcome is final
        return transaction_response

    last_ledger_sequence = result["LastLedgerSequence"]
    latest_ledger_sequence = await get_latest_validated_ledger_sequence(client)

    if last_ledger_sequence > latest_ledger_sequence:
        # outcome is not yet final
        return await _wait_for_final_transaction_outcome(
            transaction_hash, client, prelim_result, 0
        )

    raise XRPLReliableSubmissionException(
        f"The latest ledger sequence {latest_ledger_sequence} is greater than the "
        f"last ledger sequence {last_ledger_sequence} in the transaction. Prelim "
        f"result: {prelim_result}"
    )


async def send_reliable_submission(
    transaction: Transaction, client: Client
) -> Response:
    """
    Asynchronously submits a transaction and verifies that it has been included in a
    validated ledger (or has errored/will not be included for some reason).

    `See Reliable Transaction Submission
    <https://xrpl.org/reliable-transaction-submission.html>`_

    Note: This cannot be used with a standalone rippled node, because ledgers do not
    close automatically.

    Args:
        transaction: the signed transaction to submit to the ledger. Requires a
            `last_ledger_sequence` param.
        client: the network client used to submit the transaction to a rippled node.

    Returns:
        The response from a validated ledger.

    Raises:
        XRPLReliableSubmissionException: if the transaction fails, is malformed, or is
            missing a `last_ledger_sequence` param.
    """
    if transaction.last_ledger_sequence is None:
        raise XRPLReliableSubmissionException(
            "Transaction must have a `last_ledger_sequence` param."
        )
    transaction_hash = transaction.get_hash()
    submit_response = await submit(transaction, client)
    prelim_result = submit_response.result["engine_result"]
    if prelim_result[0:3] == "tem":
        raise XRPLReliableSubmissionException(
            submit_response.result["engine_result_message"]
        )

    return await _wait_for_final_transaction_outcome(
        transaction_hash, client, prelim_result, 0
    )



"""Model for AMMDeposit transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from xrpl.models.amounts import Amount, IssuedCurrencyAmount
from xrpl.models.currencies import Currency
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class AMMDepositFlag(int, Enum):
    """
    Transactions of the AMMDeposit type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_LP_TOKEN = 0x00010000
    TF_SINGLE_ASSET = 0x00080000
    TF_TWO_ASSET = 0x00100000
    TF_ONE_ASSET_LP_TOKEN = 0x00200000
    TF_LIMIT_LP_TOKEN = 0x00400000


class AMMDepositFlagInterface(FlagInterface):
    """
    Transactions of the AMMDeposit type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_LP_TOKEN: bool
    TF_SINGLE_ASSET: bool
    TF_TWO_ASSET: bool
    TF_ONE_ASSET_LP_TOKEN: bool
    TF_LIMIT_LP_TOKEN: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMDeposit(Transaction):
    """
    AMMDeposit is the deposit transaction used to add liquidity to the AMM instance
    pool, thus obtaining some share of the instance's pools in the form of LPToken.

    The following are the recommended valid combinations:
    - LPTokenOut
    - Amount
    - Amount and Amount2
    - Amount and LPTokenOut
    - Amount and EPrice
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance.
    """

    lp_token_out: Optional[IssuedCurrencyAmount] = None
    """
    Specifies the amount of shares of the AMM instance pools that the trader
    wants to redeem or trade in.
    """

    amount: Optional[Amount] = None
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance to
    deposit more of its value.
    """

    amount2: Optional[Amount] = None
    """
    Specifies the other pool asset of the AMM instance to deposit more of its
    value.
    """

    e_price: Optional[Amount] = None
    """
    Specifies the maximum effective-price that LPToken can be traded out.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_DEPOSIT,
        init=False,
    )

    def _get_errors(self: AMMDeposit) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.amount2 is not None and self.amount is None:
            errors["AMMDeposit"] = "Must set `amount` with `amount2`"
        elif self.e_price is not None and self.amount is None:
            errors["AMMDeposit"] = "Must set `amount` with `e_price`"
        elif self.lp_token_out is None and self.amount is None:
            errors["AMMDeposit"] = "Must set at least `lp_token_out` or `amount`"
        return errors



"""Codec for serializing and deserializing bridge fields."""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Type, Union

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.account_id import AccountID
from xrpl.core.binarycodec.types.issue import Issue
from xrpl.core.binarycodec.types.serialized_type import SerializedType

_TYPE_ORDER: List[Tuple[str, Type[SerializedType]]] = [
    ("LockingChainDoor", AccountID),
    ("LockingChainIssue", Issue),
    ("IssuingChainDoor", AccountID),
    ("IssuingChainIssue", Issue),
]

_TYPE_KEYS = {type[0] for type in _TYPE_ORDER}


class XChainBridge(SerializedType):
    """Codec for serializing and deserializing bridge fields."""

    def __init__(self: XChainBridge, buffer: bytes) -> None:
        """Construct a XChainBridge from given bytes."""
        super().__init__(buffer)

    @classmethod
    def from_value(
        cls: Type[XChainBridge], value: Union[str, Dict[str, str]]
    ) -> XChainBridge:
        """
        Construct a XChainBridge object from a dictionary representation of a bridge.

        Args:
            value: The dictionary to construct a XChainBridge object from.

        Returns:
            A XChainBridge object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the XChainBridge representation is invalid.
        """
        if isinstance(value, dict) and set(value.keys()) == _TYPE_KEYS:
            buffer = b""
            for (name, object_type) in _TYPE_ORDER:
                obj = object_type.from_value(value[name])
                if object_type == AccountID:
                    buffer += bytes.fromhex("14")  # AccountID length (I think)
                buffer += bytes(obj)
            return cls(buffer)

        raise XRPLBinaryCodecException(
            "Invalid type to construct a XChainBridge: expected dict,"
            f" received {value.__class__.__name__}."
        )

    @classmethod
    def from_parser(
        cls: Type[XChainBridge], parser: BinaryParser, length_hint: Optional[int] = None
    ) -> XChainBridge:
        """
        Construct a XChainBridge object from an existing BinaryParser.

        Args:
            parser: The parser to construct the XChainBridge object from.
            length_hint: The number of bytes to consume from the parser.

        Returns:
            The XChainBridge object constructed from a parser.
        """
        buffer = b""

        for (name, object_type) in _TYPE_ORDER:
            if object_type == AccountID:
                parser.skip(1)
                buffer += bytes.fromhex("14")
            obj = object_type.from_parser(parser, length_hint)
            buffer += bytes(obj)

        return cls(buffer)

    def to_json(self: XChainBridge) -> Union[str, Dict[Any, Any]]:
        """
        Returns the JSON representation of a bridge.

        Returns:
            The JSON representation of a XChainBridge.
        """
        parser = BinaryParser(str(self))
        return_json = {}
        for (name, object_type) in _TYPE_ORDER:
            if object_type == AccountID:
                parser.skip(1)
            obj = object_type.from_parser(parser, None)
            return_json[name] = obj.to_json()
        return return_json



"""Model for AMMCreate transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from typing_extensions import Final

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

AMM_MAX_TRADING_FEE: Final[int] = 1000


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMCreate(Transaction):
    """
    AMMCreate is used to create AccountRoot and the corresponding AMM
    ledger entries.
    This allows for the creation of only one AMM instance per unique asset pair.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    This field is required.
    """

    amount2: Amount = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance. This field is required.
    """

    trading_fee: int = REQUIRED  # type: ignore
    """
    Specifies the fee, in basis point, to be charged
    to the traders for the trades executed against the AMM instance.
    Trading fee is a percentage of the trading volume.
    Valid values for this field are between 0 and 1000 inclusive.
    A value of 1 is equivalent to 1/10 bps or 0.001%, allowing trading fee
    between 0% and 1%. This field is required.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_CREATE,
        init=False,
    )

    def _get_errors(self: AMMCreate) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "trading_fee": self._get_trading_fee_error(),
            }.items()
            if value is not None
        }

    def _get_trading_fee_error(self: AMMCreate) -> Optional[str]:
        if self.trading_fee < 0 or self.trading_fee > AMM_MAX_TRADING_FEE:
            return f"Must be between 0 and {AMM_MAX_TRADING_FEE}"
        return None
